require 'timeout'

module UU
  module OS
    module Lang

      # Object which is returned as result of all asynchronous tasks.
      class Future

        # Creates new instance of future object.
        # @param [#get] actor Object responsible for obtaining actual result.
        #    Accepted is any object with declared #get method.
        def initialize(actor)
          if actor.respond_to?(:get)
            @actor = actor
          elsif actor.nil?
            raise ArgumentError, 'cannot get result of asynchronous task without actor.'
            @actor = nil
          else
            raise TypeError, "cannot handle actor of type #{actor.class}"
          end
        end

        # Returns result of asynchronous task. This call is blocking (it waits until result
        # is provided or timeout exceeded).
        # @param [Fixnum] timeout Allows to configure maximum wait time within which result
        #   must be returned (defaults to 30 seconds, which is also highest possible timeout).
        # @return [Object] Asynchronous task result. Actual type of returned value depends
        #   on particular actor given in constructor.
        # @raise [Timeout::Error] In case of timeout expiration. Not that raised exception
        #   does not mean that asynchronous task was cancelled. It may or may not finish
        #   based on asynchronous task type (actual behavior depends on task implementation
        #   and cannot be affected).
        def get(timeout = 30)
          return nil if @actor.nil?
          timeout = 30 if timeout > 30
          timeout = 0.1 if timeout <= 0
          Timeout::timeout(timeout) do
            @actor.get
          end
        end

        # JSON serialization is delegated to particular actor
        # (so the result might be e.g. task URI in case of
        # asynchronous command invocation).
        # @private
        def to_json(*args) # :nodoc:
          @actor.to_json(*args)
        end

      end

    end
  end
end
