module UU
  module OS
    module Util

      # Module encapsulating methods for message formatting.
      # @private
      module MessageFormatter

        # Formatting tag using named parameter. Matching this value returns
        # two parts, first is whole tag, second is actual key.
        # @private
        NAMED_FORMAT = '([\\\%]?[\#%\$]\{([a-zA-Z0-9_\-]+)\})' # :nodoc:
        private_constant :NAMED_FORMAT

        # Formatting tag using java message format. Matching this value returns
        # two parts, first is whole tag, second is actual index.
        # @private
        JAVA_MSG_FORMAT = '([\']?\{[\']?([0-9]+)[^\}]*[\']?\}[\']?)' # :nodoc:
        private_constant :JAVA_MSG_FORMAT

        # Formatting tag using printf format. Matching this value returns
        # two parts, first is whole tag, second is formatting instruction.
        # @private
        PRINTF_FORMAT = '([\\\%]?%([^ \{]*))' # :nodoc:
        private_constant :PRINTF_FORMAT

        # Formats message using various parameters format. In case message
        # contains formatting tag which is invalid or does not match parameters,
        # tag is silently skipped (therefore this method is safe to be used for
        # logging or exception messages where unexpected formatting problem can
        # hide original message).
        # @param [String] message Message to be formatted.
        # @param [Hash, Array] parameters Message parameters
        def format_message(message, *parameters)
          return MessageFormatter.format_message(message, *parameters)
        end

        # @see #format_message
        # @private
        def self.format_message(message, *parameters) # :nodoc:
          return message if message.nil? || message.empty?
          parameters ||= []
          if parameters.size > 1 || !parameters[0].kind_of?(Hash)
            # Transform Array to Hash
            parameters = parameters.each_with_index.inject({}) do |tmp, pair|
              value, index = pair
              tmp[index] = value
              tmp
            end
          elsif parameters[0].kind_of?Hash
            parameters = parameters[0]
          else
            parameters = {}
          end

          buf = ''
          printf_index = 0
          message.scan(/#{NAMED_FORMAT}|#{JAVA_MSG_FORMAT}|#{PRINTF_FORMAT}/) do |match|
            token = match[0] || match[2] || match[4]
            parts = message.split(token, 2)
            buf << parts[0]
            message = parts[1]
            if !match[0].nil?
              buf << named_format(token, match[1], parameters)
            elsif !match[2].nil?
              buf << java_msg_format(token, match[3].to_i, parameters.values)
            elsif !match[4].nil?
              buf << printf_format(token, printf_index, parameters.values)
              printf_index += 1 if token !~ /^(\\|%%)/
            else
              buf << token
            end
          end
          return "#{buf}#{message}"
        end

        # Replaces given token by corresponding values from parameters.
        # @param [String] token Formatting token.
        # @param [String] key Internal token key to lookup value from parameters.
        # @param [Hash] parameters Message parameters.
        # @return [String] Corresponding value from parameters or original token
        #    if formatting cannot be done.
        # @private
        def self.named_format(token, key, parameters = {}) # :nodoc:
          # If escaped, remove escaping and return token
          return token[1..-1] if token =~ /^(\\|%%)/
          # Return value from parameters or original token
          value = (parameters[key] || parameters[key.to_sym] rescue nil)
          return value.nil? ? token : "#{value}"
        end
        private_class_method :named_format

        # Replaces given token by corresponding values from parameters.
        # @param [String] token Formatting token.
        # @param [Fixnum] index Parameter value index.
        # @param [Array] parameters Message parameters.
        # @return [String] Corresponding value from parameters or original token
        #    if formatting cannot be done.
        # @private
        def self.java_msg_format(token, index, parameters = []) # :nodoc:
          # If escaped, remove escaping and return token
          if token =~ /(^'([^']+)'$)|(^'\{')|('\}'$)/
            return token.sub(/'\{'/, '{').sub(/'\}'/, '}').gsub(/(^')|('$)/, '')
          end
          # Not enough parameters, return original token
          return token if index >= parameters.size
          # Return value from parameters
          return "#{parameters[index]}"
        end
        private_class_method :java_msg_format

        # Replaces given token by corresponding values from parameters.
        # @param [String] token Formatting token.
        # @param [Fixnum] index Parameter value index.
        # @param [Array] parameters Message parameters.
        # @return [String] Corresponding value from parameters or original token
        #    if formatting cannot be done.
        # @private
        def self.printf_format(token, index, parameters = []) # :nodoc:
          # If escaped, remove escaping and return token
          return token[1..-1] if token =~ /^(\\|%%)/
          # Not enough parameters, return original token
          return token if index >= parameters.size
          # Try to format or return original token
          return (token % parameters[index] rescue token)
        end
        private_class_method :printf_format

      end

    end
  end
end
